from typing import Any


class InputSession:
    def __init__(self, manual_delete: bool = False):
        self.messages = []
        self.manual_delete = manual_delete
        self.directory_level: int = None
    
    def append(self, message: Any):
        self.messages.append(message)


from types import CoroutineType
from typing import Any, Callable
from abc import ABC, abstractmethod

class CallbackData(ABC):
    @abstractmethod
    def clone(self): ...

class Dummy(CallbackData):
    def clone(self):
        return Dummy()

class RunFunc(CallbackData):
    def __init__(self, function: Callable, **kwargs):
        """Использование:  
            function - Функция для выполнения при нажатии кнопки  
            **kwargs - keyword аргументы функции
        """
        assert isinstance(function, Callable)
        self.function = function
        self.kwargs = kwargs
    
    def clone(self):
        return RunFunc(self.function, **self.kwargs)
    
    def __eq__(self, other: "RunFunc"):
        return isinstance(other, RunFunc) and \
            self.function == other.function and self.kwargs == other.kwargs

class GoToScreen(CallbackData):
    def __init__(self, screen_name: str, *
            , pre_func: Callable[ [int], None ] = None
            , post_func: Callable[ [int], None ] = None):
        assert isinstance(screen_name, str)
        assert pre_func is None or isinstance(pre_func, Callable)
        assert post_func is None or isinstance(post_func, Callable)
        
        self.screen_name = screen_name
        self.pre_func = pre_func
        self.post_func = post_func
    
    def clone(self):
        return GoToScreen(self.screen_name)
    
    def __eq__(self, other: "GoToScreen"):
        return isinstance(other, GoToScreen) and \
            self.screen_name == other.screen_name

class StepBack(CallbackData):
    def __init__(self, times: int = 1, clear_input_callback: bool = True
            , pop_last_input: bool = True
            , pre_func: Callable[ [int], None ] = None
            , post_func: Callable[ [int], None ] = None):
        self.times = times
        self.clear_input_callback = clear_input_callback
        self.pop_last_input = pop_last_input
        self.pre_func = pre_func
        self.post_func = post_func

    def clone(self):
        return StepBack()
    
    def __eq__(self, other: "StepBack"):
        return isinstance(other, StepBack)

class CallbackDataMapping:
    def __init__(self):
        self.items = []
    
    def add(self, callback: CallbackData, uuid: str):
        self.items.append((callback, uuid))
    
    def get_by_callback(self, callback: CallbackData):
        for i_callback, uuid in self.items:
            if callback == i_callback:
                return uuid
        raise KeyError(callback)
    
    def get_by_uuid(self, uuid: str):
        for callback, i_uuid in self.items:
            if uuid == i_uuid:
                return callback
        return None


from .input_session import InputSession
from .input_callback import InputCallback
from .callback_data import CallbackData, CallbackDataMapping
from .screen import ReadyScreen, SentScreen

class UserData:
    def __init__(self, user_id: int):
        self.user_id = user_id
        self.callback_mapping = CallbackDataMapping()
        self.media_group_id: str = None
        self.input_callback: InputCallback = None
        self.directory_stack: list[str] = []
        self.screen: SentScreen = None
        self.screen_buffer: ReadyScreen = None
        self.__input_session: InputSession = None
    
    @property
    def input_session(self):
        return self.__input_session
    
    @input_session.setter
    def input_session(self, value: InputSession):
        self.__input_session = value
        if value and not value.manual_delete:
            value.directory_level = len(self.directory_stack)
    
    def update_input_session(self):
        new_dir_level = len(self.directory_stack)
        ses = self.__input_session
        if ses and not ses.manual_delete and ses.directory_level > new_dir_level:
            self.__input_session = None

class UserDataManager:
    def __init__(self):
        self.__users_data: dict[int, UserData] = {}
        self.users_data = self.__users_data
    
    def get(self, user_id: int) -> UserData:
        user_data = self.__users_data.get(user_id)
        if user_data is None:
            user_data = UserData(user_id)
            self.set(user_id, user_data)
        return user_data
    
    def reset(self, user_id: int) -> None:
        self.set(user_id, UserData(user_id))
    
    def set(self, user_id: int, user_data: UserData):
        self.__users_data[user_id] = user_data



from typing import Any
from ..input_session import InputSession as BaseInputSession
from telegram import Message as TgMessage


class InputSession(BaseInputSession):
    def __init__(self, manual_delete: bool = False):
        super().__init__(manual_delete)
        self.messages: list[TgMessage] = []
    
    def append(self, message: TgMessage):
        self.messages.append(message)




from ..user_data import UserData as BaseUserData, \
    UserDataManager as BaseUserDataManager
from .screen import SentScreen
from .input_session import InputSession



class UserData(BaseUserData):
    def __init__(self, user_id: int):
        super().__init__(self, user_id)
        self.screen: SentScreen = None
        self.__input_session: InputSession = None
    
    @property
    def input_session(self):
        return self.__input_session
    
    @input_session.setter
    def input_session(self, value: InputSession):
        super().input_session = value
        if value is None:
            print("input session deleted")
        

class UserDataManager(BaseUserDataManager):
    def __init__(self):
        self.__users_data: dict[int, UserData] = {}
        self.users_data = self.__users_data
    
    def get(self, user_id: int) -> UserData:
        user_data = self.__users_data.get(user_id)
        if user_data is None:
            user_data = UserData(user_id)
            self.set(user_id, user_data)
        return user_data
    
    def reset(self, user_id: int) -> None:
        self.set(user_id, UserData(user_id))
    
    def set(self, user_id: int, user_data: UserData):
        self.__users_data[user_id] = user_data
